#!/usr/bin/env python
"""
Backport pull requests to a particular branch.

Usage: backport_pr.py branch [PR] [PR2]

e.g.:

    python tools/backport_pr.py 0.13.1 123 155

to backport PR #123 onto branch 0.13.1

or
    python tools/backport_pr.py 2.1

to see what PRs are marked for backport with milestone=2.1 that have yet to be applied
to branch 2.x.

Forked from the backport_pr.py script in the ipython/ipython repository.

"""

from __future__ import print_function

import os
import re
import sys

from subprocess import Popen, PIPE, check_call, check_output
try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from gh_api import (
    get_pull_request,
    get_pull_request_files,
    get_issues_list,
    get_milestone_id,
    is_pull_request
)


def find_rejects(root='.'):
    for dirname, dirs, files in os.walk(root):
        for fname in files:
            if fname.endswith('.rej'):
                yield os.path.join(dirname, fname)


def get_current_branch():
    branches = check_output(['git', 'branch'])
    for branch in branches.splitlines():
        if branch.startswith(b'*'):
            return branch[1:].strip().decode('utf-8')


def backport_pr(branch, num, project='jupyter/nbgrader'):
    current_branch = get_current_branch()
    if branch != current_branch:
        check_call(['git', 'checkout', branch])
    check_call(['git', 'pull'])
    pr = get_pull_request(project, num, auth=False)
    files = get_pull_request_files(project, num, auth=False)
    patch_url = pr['patch_url']
    title = pr['title']
    description = pr['body']
    fname = "PR%i.patch" % num
    if os.path.exists(fname):
        print("using patch from {fname}".format(**locals()))
        with open(fname, 'rb') as f:
            patch = f.read()
    else:
        req = urlopen(patch_url)
        patch = req.read()

    lines = description.splitlines()
    if len(lines) > 5:
        lines = lines[:5] + ['...']
        description = '\n'.join(lines)

    msg = "Backport PR #%i: %s" % (num, title) + '\n\n' + description
    check = Popen(['git', 'apply', '--check', '--verbose'], stdin=PIPE)
    check.communicate(patch)

    if check.returncode:
        print("patch did not apply, saving to {fname}".format(**locals()))
        print("edit {fname} until `cat {fname} | git apply --check` succeeds".format(**locals()))
        print("then run tools/backport_pr.py {num} again".format(**locals()))
        if not os.path.exists(fname):
            with open(fname, 'wb') as f:
                f.write(patch)
        return 1

    p = Popen(['git', 'apply'], stdin=PIPE)
    p.communicate(patch)

    filenames = [f['filename'] for f in files]

    check_call(['git', 'add'] + filenames)
    check_call(['git', 'commit', '-m', msg])

    print("PR #%i applied, with msg:" % num)
    print()
    print(msg)
    print()

    if branch != current_branch:
        check_call(['git', 'checkout', current_branch])

    return 0

backport_re = re.compile(r"(?:[Bb]ackport|[Mm]erge).*#(\d+)")

def already_backported(branch, since_tag=None):
    """return set of PRs that have been backported already"""
    if since_tag is None:
        since_tag = check_output(['git', 'describe', branch, '--abbrev=0', '--tags']).decode('utf8').strip()
    cmd = ['git', 'log', '%s..%s' % (since_tag, branch), '--oneline']
    lines = check_output(cmd).decode('utf8')
    return set(int(num) for num in backport_re.findall(lines))

def should_backport(labels=None, milestone=None):
    """return set of PRs marked for backport"""
    if labels is None and milestone is None:
        raise ValueError("Specify one of labels or milestone.")
    elif labels is not None and milestone is not None:
        raise ValueError("Specify only one of labels or milestone.")
    if labels is not None:
        issues = get_issues_list(
            "jupyter/nbgrader",
            labels=labels,
            state='closed',
            auth=False)
    else:
        milestone_id = get_milestone_id(
            "jupyter/nbgrader",
            milestone,
            auth=False)
        issues = get_issues_list(
            "jupyter/nbgrader",
            milestone=milestone_id,
            state='closed',
            auth=False)

    should_backport = set()
    for issue in issues:
        if not is_pull_request(issue):
            continue
        pr = get_pull_request(
            "jupyter/nbgrader",
            issue['number'],
            auth=False)
        if not pr['merged']:
            print ("Marked PR closed without merge: %i" % pr['number'])
            continue
        if pr['base']['ref'] != 'master':
            continue
        should_backport.add(pr['number'])
    return should_backport

if __name__ == '__main__':

    if len(sys.argv) < 2:
        print(__doc__)
        sys.exit(1)

    if len(sys.argv) < 3:
        milestone = sys.argv[1]
        branch = ".".join(milestone.split('.')[:-1]) + '.x'
        already = already_backported(branch)
        should = should_backport(milestone=milestone)
        print ("The following PRs should be backported:")
        for pr in sorted(should.difference(already)):
            print (pr)
        sys.exit(0)

    for prno in map(int, sys.argv[2:]):
        print("Backporting PR #%i" % prno)
        rc = backport_pr(sys.argv[1], prno)
        if rc:
            print("Backporting PR #%i failed" % prno)
            sys.exit(rc)



#!/usr/bin/env python
"""Print a list of merged PRs for a given milestone.

Usage:

    python tools/changelog.py MILESTONE

"""

import sys
from gh_api import (
    get_milestone_id,
    get_pulls_list
)

if __name__ == "__main__":

    if len(sys.argv) != 2:
        print(__doc__)
        sys.exit(1)

    milestone = sys.argv[1]

    all_pulls = get_pulls_list(
        "jupyter/nbgrader",
        state='closed',
        auth=True)

    pulls = []
    for pr in all_pulls:
        if pr['milestone']['title'] == milestone and pr['merged_at']:
            pulls.append(pr)

    print()
    print("The following PRs were merged for the {} milestone:".format(milestone))
    print("-----------------------------------------------------")
    for pull in pulls:
        print("- PR #{}: {}".format(pull['number'], pull['title']))
    print("-----------------------------------------------------")



#!/usr/bin/env python

import os
import sys
import subprocess as sp
import argparse


def echo(msg):
    print("\033[1;37m{0}\033[0m".format(msg))


def run(cmd):
    echo(cmd)
    return sp.check_call(cmd, shell=True)


try:
    from nbformat import read
except ImportError:
    echo("Warning: nbformat could not be imported, some tasks may not work")

WINDOWS = sys.platform == 'win32'


def _check_if_directory_in_path(pth, target):
    while pth != '':
        pth, dirname = os.path.split(pth)
        if dirname == target:
            return True
    return False


def docs(args):
    del args  # unused
    run('git clean -fdX nbgrader/docs')
    if not WINDOWS:
        run('pytest --nbval-lax --current-env nbgrader/docs/source/user_guide/*.ipynb')
    run('python nbgrader/docs/source/build_docs.py')
    run('python nbgrader/docs/source/clear_docs.py')
    run('make -C nbgrader/docs html')
    run('make -C nbgrader/docs linkcheck')


def cleandocs(args):
    del args  # unused
    run('python nbgrader/docs/source/clear_docs.py')


def _run_tests(mark, skip, junitxml, paralell=False):
    cmd = []
    cmd.append('pytest')
    if not WINDOWS:
        cmd.append('--cov nbgrader')
        cmd.append('--no-cov-on-fail')
    if junitxml:
        cmd.extend(['--junitxml', junitxml])
    cmd.append('-v')
    cmd.append('-x')
    if paralell:
        cmd.extend(['--numprocesses', 'auto'])
    cmd.extend(['--reruns', '4'])

    marks = []
    if mark is not None:
        marks.append(mark)
    if skip is not None:
        marks.append("not {}".format(skip))
    if len(marks) > 0:
        cmd.append('-m "{}"'.format(" and ".join(marks)))

    run(" ".join(cmd))

    if not WINDOWS:
        run("ls -a .coverage*")
        run("coverage combine || true")


def tests(args):
    if args.group == 'python':
        _run_tests(
            mark="not nbextensions", skip=args.skip, junitxml=args.junitxml, paralell=True)

    elif args.group == 'nbextensions':
        _run_tests(mark="nbextensions", skip=args.skip, junitxml=args.junitxml)

    elif args.group == 'docs':
        docs(args)

    elif args.group == 'all':
        _run_tests(mark=None, skip=args.skip, junitxml=args.junitxml)

    else:
        raise ValueError("Invalid test group: {}".format(args.group))


def aftersuccess(args):
    if args.group in ('python', 'nbextensions'):
        run('codecov')
    else:
        echo('Nothing to do.')


def js(args):
    run('npm install')
    run('./node_modules/.bin/bower install --config.interactive=false')
    if args.clean:
        run('git clean -fdX nbgrader/server_extensions/formgrader/static/components')


def install(args):
    # The docs don't seem to build correctly if it's a symlinked install.
    if args.group == 'docs':
        cmd = 'pip install -r dev-requirements.txt .'
    else:
        cmd = 'pip install -r dev-requirements.txt -e .'
    run(cmd)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    # docs
    docs_parser = subparsers.add_parser('docs')
    docs_parser.set_defaults(func=docs)

    # cleandocs
    cleandocs_parser = subparsers.add_parser('cleandocs')
    cleandocs_parser.set_defaults(func=cleandocs)

    # tests
    tests_parser = subparsers.add_parser('tests')
    tests_parser.add_argument('--group', type=str, default='all')
    tests_parser.add_argument('--skip', type=str, default=None)
    tests_parser.add_argument('--junitxml', type=str, default=None)
    tests_parser.set_defaults(func=tests)

    # aftersuccess
    aftersuccess_parser = subparsers.add_parser('aftersuccess')
    aftersuccess_parser.add_argument('--group', type=str, required=True)
    aftersuccess_parser.set_defaults(func=aftersuccess)

    # js
    js_parser = subparsers.add_parser('js')
    js_parser.add_argument('--clean', type=bool, default=True)
    js_parser.set_defaults(func=js)

    # install
    install_parser = subparsers.add_parser('install')
    install_parser.add_argument('--group', type=str, required=True)
    install_parser.set_defaults(func=install)

    args = parser.parse_args()
    args.func(args)



#!/usr/bin/env python
"""Print a list of contributors for a particular milestone.

Usage:

    python tools/contributor_list.py [MILESTONE] [MILESTONE] ...

"""

import sys
from gh_api import (
    get_milestone,
    get_milestones,
    get_milestone_id,
    get_issues_list,
)

if __name__ == "__main__":

    if len(sys.argv) < 2:
        milestones = get_milestones("jupyter/nbgrader", auth=True)
    else:
        milestones = [get_milestone("jupyter/nbgrader", milestone_id, auth=True)
                      for milestone_id in sys.argv[1:]]

    users = set()
    for milestone in milestones:
        if milestone['title'] == "No action":
            continue

        print("Getting users for {}...".format(milestone['title']))

        # this returns both issues and PRs
        issues = get_issues_list(
            "jupyter/nbgrader",
            state='all',
            milestone=milestone['number'],
            auth=True)

        for issue in issues:
            users.add(issue['user']['login'])

    users = {user.lower(): user for user in users}
    print()
    print("The following users have submitted issues and/or PRs:")
    print("-----------------------------------------------------")
    for user in sorted(users.keys()):
        print("- {}".format(users[user]))
    print("-----------------------------------------------------")



c = get_config()

# Our user list
c.Authenticator.whitelist = [
    'instructor1',
    'instructor2',
    'student1',
    'grader-course101',
    'grader-course123'
]

c.Authenticator.admin_users = {
    'instructor1',
    'instructor2'
}

# instructor1 and instructor2 have access to different shared servers:
c.JupyterHub.load_groups = {
    'formgrade-course101': [
        'instructor1',
        'grader-course101',
    ],
    'formgrade-course123': [
        'instructor2',
        'grader-course123'
    ],
    'nbgrader-course101': [],
    'nbgrader-course123': []
}

# Start the notebook server as a service. The port can be whatever you want
# and the group has to match the name of the group defined above.
c.JupyterHub.services = [
    {
        'name': 'course101',
        'url': 'http://127.0.0.1:9999',
        'command': [
            'jupyterhub-singleuser',
            '--group=formgrade-course101',
            '--debug',
        ],
        'user': 'grader-course101',
        'cwd': '/home/grader-course101',
        'api_token': '{{course101_token}}'
    },
    {
        'name': 'course123',
        'url': 'http://127.0.0.1:9998',
        'command': [
            'jupyterhub-singleuser',
            '--group=formgrade-course123',
            '--debug',
        ],
        'user': 'grader-course123',
        'cwd': '/home/grader-course123',
        'api_token': '{{course123_token}}'
    },
]

